/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * See LICENSE.txt included in this distribution for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at LICENSE.txt.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
 * Portions Copyright 2012 Jens Elkner.
 */

package org.opensolaris.opengrok.index;

import java.io.File;
import java.io.Serializable;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A file pattern filter.
 * 
 * @version	$Revision$
 */
public class Filter implements Serializable {
    private static final long serialVersionUID = 3L;

    /** The list of exact filenames */
    private final HashSet<String> filenames;
    /** The list of filenames with wildcards */
    private final LinkedHashSet<Pattern> patterns;
    /** The list of full paths */
    private final HashSet<String> paths;
    /**
     * The full list of all patterns. This list will be saved in the
     * configuration file (if used)
     */
    private final LinkedHashSet<String> items;

    /**
     * Creates a new instance which matches nothing (no patterns).
     */
    public Filter() {
        this(null);
    }

    /**
     * Creates a new instance which matches the given patterns.
     * @param patterns patterns to add to this instance.
     */
    public Filter(String[] patterns) {
        filenames = new LinkedHashSet<String>();
        this.patterns = new LinkedHashSet<Pattern>();
        paths = new LinkedHashSet<String>();
        items = new LinkedHashSet<String>();
        if (patterns != null) {
            for (String s : patterns) {
                add(s);
            }
        }
    }

    /**
     * Check, whether this instance contains any items, which could be matched.
     * @return {@code true} if it contains any items.
     */
    public boolean isEmpty() {
        return items.isEmpty();
    }

    /**
     * Get the complete list of items that would be matched by this matcher.
     * @return a list of all wildcards, exact lists and paths that this filter
     *         contains.
     */
    public LinkedHashSet<String> getItems() {
        return items;
    }

    /**
     * Specify a new filter to use.
     * @param item the new filter.
     */
    public void setItems(LinkedHashSet<String> item) {
        clear();
        for (String s : item) {
            add(s);
        }
    }

    /**
     * Add a pattern to the list of patterns.
     * @param pattern the pattern to filename.
     */
    public void add(String pattern) {
        if (items.add(pattern)) {
            addPattern(pattern);
        }
    }

    /**
     * Remove all installed patterns from this instance.
     */
    public void clear() {
        patterns.clear();
        filenames.clear();
        paths.clear();
        items.clear();
    }

    /**
     * Check, whether the given path's basename, its absolute name or extension 
     * matches any pattern.
     * @param path the path to check.
     * @return {@code true} if a match was found.
     */
    public boolean match(File path) {
        String fileName = path.getName();
        if (filenames.contains(fileName)) {
            return true;
        }
        for (Pattern p : patterns) {
            Matcher m = p.matcher(fileName);
            if (m.matches()) {
                return true;
            }
        }
        String absolute = path.getAbsolutePath();
        for (String s : paths) {
            if (absolute.endsWith(s)) {
                return true;
            }
        }
        //Check File extension
        int start = fileName.indexOf(".");
        if (start != -1){
            String fileExtension = fileName.substring(start);
            if (filenames.contains(fileExtension)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Check, whether the given path's basename, its absolute name or extension 
     * matches any pattern.
     * @param path the path to check.
     * @return {@code true} if a match has been found.
     */
    public boolean match(String path) {
        return match(new File(path));
    }

    /**
     * Add a pattern to the correct list of internal filters to match.
     *
     * @param pattern the pattern to add
     */
    private void addPattern(String pattern) {
        if (pattern.contains("*") || pattern.contains("?")) {
            patterns.add(compilePattern(pattern));
        } else if (pattern.contains(File.separator)) {
            if (pattern.charAt(0) == File.separatorChar) {
                paths.add(pattern);
            } else {
                paths.add(File.separator + pattern);
            }
        } else {
            filenames.add(pattern);
        }
    }

    /**
     * Convert the glob pattern (examples: *.c, *.?xx) to a regular expression
     * and compile it.
     *
     * @param pattern a pattern to match file names against
     * @return a compiled regular expression representing the pattern
     */
    private static Pattern compilePattern(String pattern) {
        // Build the regex by replacing "*" with ".*" and "?" with ".". All
        // other characters should be quoted to ensure exact match.
        StringBuilder regex = new StringBuilder();
        int pos = 0;
        String[] components = pattern.split("[*?]");
        for (String str : components) {
            if (str.length() > 0) {
                // Quote the characters up to next wildcard or end of string.
                regex.append(Pattern.quote(str));
                pos += str.length();
            }
            if (pos < pattern.length()) {
                // Replace wildcard with equivalent regular expression.
                if (pattern.charAt(pos) == '*') {
                    regex.append(".*");
                } else {
                    assert pattern.charAt(pos) == '?';
                    regex.append('.');
                }
                pos++;
            }
        }

        // Compile the regex.
        return Pattern.compile(regex.toString());
    }
}
