/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * See LICENSE.txt included in this distribution for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at LICENSE.txt.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright (c) 2007, 2010, Oracle and/or its affiliates. All rights reserved.
 * Portions Copyright 2012 Jens Elkner.
 */
package org.opensolaris.opengrok.web;

import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.net.UnknownHostException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;
import javax.servlet.ServletRequestEvent;
import javax.servlet.ServletRequestListener;

import org.opensolaris.opengrok.configuration.RuntimeEnvironment;

/**
 * Listener to manage the runtime environment.
 * @author Trond Norbye
 */
public final class WebappListener 
    implements ServletContextListener, ServletRequestListener
{
    /** Boolean servlet parameter to indicate, whether to serve debug enabled JS */ 
    public static final String JS_DEBUG_PARAM = "JsDebug";

    private static final Logger logger = Logger.getLogger(WebappListener.class.getName());
    /**
     * {@inheritDoc}
     */
    @Override
    public void contextInitialized(final ServletContextEvent servletContextEvent) {
        ServletContext context = servletContextEvent.getServletContext();

        String config = context.getInitParameter("CONFIGURATION");
        if (config == null) {
            logger.severe("CONFIGURATION section missing in web.xml");
        } else {
            try {
                RuntimeEnvironment.readConfig(new File(config));
            } catch (IOException ex) {
                logger.warning("OpenGrok Configuration error. Failed to read config file: " 
                    + ex.getMessage());
                logger.log(Level.FINE, "contextInitialized", ex);
            }
        }

        String address = context.getInitParameter("ConfigAddress");
        if (address != null && address.length() > 0) {
            logger.log(Level.CONFIG, "Will listen for configuration on {0}", address);
            String[] cfg = address.split(":");
            if (cfg.length == 2) {
                try {
                    SocketAddress addr = 
                        new InetSocketAddress(InetAddress.getByName(cfg[0]),
                            Integer.parseInt(cfg[1]));
                    if (!RuntimeEnvironment.startConfigListenerThread(addr)) {
                        logger.warning("Failed to start configuration listener thread");
                    }
                } catch (NumberFormatException ex) {
                    logger.warning("Failed to start configuration listener thread: " 
                        + ex.getMessage());
                    logger.log(Level.FINE, "contextInitialized", ex);
                } catch (UnknownHostException ex) {
                    logger.warning("Failed to start configuration listener thread:"
                        + ex.getMessage());
                    logger.log(Level.FINE, "contextInitialized", ex);
                }
            } else {
                StringBuilder sb = new StringBuilder("Incorrect format for the configuration address - ");
                for (int i = 0; i < cfg.length; ++i) {
                    sb.append('[').append(cfg[i]).append("], ");
                }
                sb.setLength(sb.length()-2);
                logger.warning(sb.toString());
            }
        }
        updateContext(JS_DEBUG_PARAM, context.getInitParameter(JS_DEBUG_PARAM), 
            context);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void contextDestroyed(final ServletContextEvent servletContextEvent) {
        RuntimeEnvironment.stopConfigListenerThread();
    }

    private static void updateContext(String name, Object value, ServletContext ctx) {
        if (name == null || ctx == null || !(name.equals(JS_DEBUG_PARAM))) {
            return;
        }
        if (value != null && Boolean.parseBoolean(value.toString())) {
            ctx.setAttribute(name, Boolean.TRUE);
        } else {
            ctx.setAttribute(name, Boolean.FALSE);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void requestDestroyed(ServletRequestEvent e) {
        PageConfig.cleanup(e.getServletRequest());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void requestInitialized(ServletRequestEvent e) {
        // pass through
    }
}
